//version: 1.8
using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Printing;
using System.Runtime.InteropServices;
using System.Threading;

namespace RichTextBoxDesktop
{
    //Inherit from IDisposable if you need Basic4ppc to free resources when the application is closed.
    public class RichTextBoxDesktop : IDisposable
    {
        // private variables
        private RichTextBox rtb ; //The "real" object.
        
        private Font ifont ;
        private String ifontname;
        private Color iColor;
        private int iBold; //1
        private int iItalic; //2
        private int iStrikeout; // 8
        private int iUnderline;// 4
        private float ifontsize;

        private PrintDocument pdoc;
        private PrintDialog pdial;
        private PrintPreviewDialog pprev;
        private PageSetupDialog psetup;

        private Image ClipImage;

         //constructor
        public RichTextBoxDesktop(Control FormName, int Left, int Top, int Width, int Height)
        {
            rtb = new RichTextBox();
            rtb.Left = Left;
            rtb.Top = Top;
            rtb.Width = Width;
            rtb.Height = Height;
            FormName.Controls.Add(rtb);
            SetFontDefaults();

            pdoc = new PrintDocument();
            pdial = new PrintDialog();
            pprev = new PrintPreviewDialog();
            psetup = new PageSetupDialog();

            pdial.AllowCurrentPage = true;
            pdial.AllowPrintToFile = true;
            pdial.AllowSomePages = true;

            psetup.EnableMetric = true;

            pprev.Document = pdoc;
            pdial.Document = pdoc;
            psetup.Document = pdoc;

            pdoc.BeginPrint += new System.Drawing.Printing.PrintEventHandler(pdoc_BeginPrint);
            pdoc.PrintPage += new System.Drawing.Printing.PrintPageEventHandler(pdoc_PrintPage);
           
       }

        //richtextbox properties

        // 1.1 added offset, hangingindent, leftindent, rightindent, bulleted,
        // backgroundcolor, text & rtf properties. Paste, Tab and NewPage method
        // 1.2 added printpreview properties, print and printdialog, pagesetupmetric,
        // printer in page setup
        // 1.3 added pagesetup properties
        // 1.4 added copy image to clipboard
        // 1.5 added paste text
        // 1.6 added cut and copy selection to clipboard
        // 1.7 added scrollbars and wrap properties
        // 1.8 added SelectedRtf, SelectedText, SelectionFont, SelectionStart and SelectionLength properties
        //     added GetSelectionFormat method
        //     amended Alignment property and AddText
        //     removed FontScript and FontDialog.Script properties
        //     added FontDialog.FontRef
        public double DLLVersion
        { get { return 1.8; } }

        public RichTextBox ControlRef
        { get { return rtb; } }

        public int Height
        { get { return rtb.Height; } set { rtb.Height = value; } }

        public int Width
        { get { return rtb.Width; } set { rtb.Width = value; } }

        public int Left
        { get { return rtb.Left; } set { rtb.Left = value; } }

        public int Top
        { get { return rtb.Top; } set { rtb.Top = value; } }

        public Color BackgroundColor
        { get { return rtb.BackColor; } set { rtb.BackColor = value; } }

        public bool Visible
        { get { return rtb.Visible; } set { rtb.Visible = value; } }


        // None = 0, Horizontal = 1, Vertical = 2Both = 3,  
        //ForcedHorizontal = 0x11,ForcedVertical = 0x12,  ForcedBoth = 0x13,
        public int ScrollBars
        { get { return (int) rtb.ScrollBars; } set { rtb.ScrollBars = (RichTextBoxScrollBars ) value; } }

        public bool WordWrap
        { get { return rtb.WordWrap; } set { rtb.WordWrap = value; } }

        public String Font
        { get { return ifontname; } set { ifontname = value; SetFont(); } }

        public Color Color
        { get { return rtb.SelectionColor; } set { rtb.SelectionColor = value; } }

        public Color BackColor
        { get { return rtb.SelectionBackColor; } set { rtb.SelectionBackColor = value; } }

        public int Alignment
        { get { return (int)rtb.SelectionAlignment; } set { rtb.SelectionAlignment = (HorizontalAlignment)value; } }
        //{ get { return iAlignment; } set { iAlignment = value; } }

        public bool Bold
        { get { return rtb.SelectionFont.Bold; } set { iBold = (value ? 1 : 0); SetFont(); } }
        
        public bool Italic
        { get { return rtb.SelectionFont.Italic; } set { iItalic = (value ? 2 : 0); SetFont(); } }
        
        public bool StrikeOut
        { get { return rtb.SelectionFont.Strikeout; } set { iStrikeout = (value ? 8 : 0); SetFont(); } }
       
        public bool Underline
        { get { return rtb.SelectionFont.Underline; } set { iUnderline = (value ? 4 : 0); SetFont(); } }

        public float FontSize
        { get { return rtb.SelectionFont.SizeInPoints; } set { ifontsize = value; SetFont(); } }

        public String Text
        { get { return rtb.Text; } set { rtb.Text = value; } }

        public String Rtf
        { get { return rtb.Rtf; } set { rtb.Rtf = value; } }

        public bool Bulleted
        { get { return rtb.SelectionBullet; } set { rtb.SelectionBullet = value; } }

        public int LeftIndent
        { get { return rtb.SelectionIndent; } set { rtb.SelectionIndent = value; } }

        public int RightIndent
        { get { return rtb.SelectionRightIndent; } set { rtb.SelectionRightIndent = value; } }

        public int Offset
        { get { return rtb.SelectionCharOffset; } set { rtb.SelectionCharOffset = value; } }

        public int HangingIndent
        { get { return rtb.SelectionHangingIndent; } set { rtb.SelectionHangingIndent = value; } }
 
        // print preview properties
        public int PrintPreviewHeight
        { get { return pprev.Height; } set { pprev.Height = value; } }

        public int PrintPreviewWidth
        { get { return pprev.Width; } set { pprev.Width = value; } }

        public int PrintPreviewLeft
        { get { return pprev.Left; } set { pprev.Left = value; } }

        public int PrintPreviewTop
        { get { return pprev.Top; } set { pprev.Top = value; } }

        public int PrintPreviewState
        { get { return (int)pprev.WindowState; } set { pprev.WindowState = (FormWindowState) value; } }


        // pagesetup properties
        public bool PageSetupMetric
        { get { return psetup.EnableMetric; } set { psetup.EnableMetric = value; } }

        public bool Landscape
        { get { return psetup.PageSettings.Landscape; } set { psetup.PageSettings.Landscape = value; } }

        public int LeftMargin
        { get { return psetup.PageSettings.Margins.Left; } set { psetup.PageSettings.Margins.Left = value; } }

        public int RightMargin
        { get { return psetup.PageSettings.Margins.Right; } set { psetup.PageSettings.Margins.Right = value; } }

        public int TopMargin
        { get { return psetup.PageSettings.Margins.Top; } set { psetup.PageSettings.Margins.Top = value; } }

        public int BottomMargin
        { get { return psetup.PageSettings.Margins.Bottom; } set { psetup.PageSettings.Margins.Bottom = value; } }


        // selection properties and method
        public String SelectedRtf
        { get { return rtb.SelectedRtf; } set { rtb.SelectedRtf = value; } }

        public String SelectedText
        { get { return rtb.SelectedText; } set { rtb.SelectedText = value; } }

        public Font SelectionFont
        { get { return ifont; }
          set { ifont = value;
                ifontname = ifont.Name;
                ifontsize = ifont.SizeInPoints;
                iBold = ifont.Bold ? 1 : 0;
                iItalic = ifont.Italic ? 2 : 0;
                iStrikeout = ifont.Strikeout ? 8 : 0;
                iUnderline = ifont.Underline ? 4 : 0;
                rtb.SelectionFont = ifont; } }

        public int SelectionLength
        { get { return rtb.SelectionLength; } set { rtb.SelectionLength = value; } }

        public int SelectionStart
        { get { return rtb.SelectionStart; } set { rtb.SelectionStart = value; } }

        public void GetSelectionFont()
        {
            ifont = rtb.SelectionFont;
            ifontname = ifont.Name;
            ifontsize = ifont.SizeInPoints;
            iBold = ifont.Bold ? 1 : 0;
            iItalic = ifont.Italic ? 2 : 0;
            iStrikeout = ifont.Strikeout ? 8 : 0;
            iUnderline = ifont.Underline ? 4 : 0;
        }



        // methods
        public void AddText(String txt)
        {
            //switch (iAlignment)
            //{
            //    case 1:
            //        rtb.SelectionAlignment = HorizontalAlignment.Right;
            //        break;
            //    case 2:
            //        rtb.SelectionAlignment = HorizontalAlignment.Center;
            //        break;
            //    default:
            //        rtb.SelectionAlignment = HorizontalAlignment.Left;
            //        break;
            //}            
            //FontStyle ifontstyle = (FontStyle)iBold + iStrikeout + iItalic + iUnderline;
            //ifont = new Font(ifontname, ifontsize, ifontstyle, GraphicsUnit.Point);
            //rtb.SelectionFont = ifont;
            rtb.AppendText(txt);
        }

        public void Clear()
        {
            rtb.Clear();
            SetFontDefaults();
        }

        public void CopyImageToClipBoard(Image img)
        {
            ClipImage = img;
            Thread cbThread = new Thread(new ThreadStart(CopyToClipboard));
            cbThread.ApartmentState = ApartmentState.STA;
            cbThread.Start();
            cbThread.Join();
        }

        public void CopySelectionToClipboard()
        {
            if (rtb.SelectionLength > 0)
                rtb.Copy();
        }

        public void CutSelectionToClipboard()
        {
            if (rtb.SelectionLength > 0)
                rtb.Cut();
        }
    

        public void LoadFile(String path)
        {
            rtb.LoadFile(path);
        }

        public void SaveFile(String path)
        {
            rtb.SaveFile(path);
        }

        public void Paste()
        {
            rtb.Paste();
        }

        public void PasteBitmap()
        {
            rtb.Paste(DataFormats.GetFormat(DataFormats.Bitmap));
        }

        public void PasteText()
        {
            rtb.Paste(DataFormats.GetFormat(DataFormats.Text));
        }
 
        public void Tab()
        {
            rtb.AppendText( "\t");
        }

        public void NewPage()
        {
            rtb.AppendText("\f");
        }

        // public printing methods
        public int PageSetup()
        {
            return (int)psetup.ShowDialog();
        }

        public int PrintPreview()
        {
            return (int) pprev.ShowDialog();
        }

        public int PrintDialog()
        {
            if (pdial.ShowDialog() == DialogResult.OK)
            {
                pdoc.Print();
                return (int) DialogResult.OK;
            }
            else
            {
                return (int) DialogResult.Cancel;
            }
        }

        public void Print()
        {
            pdoc.Print();
        }

        // dispose method
        public void Dispose()
        {
        }    


        //private methods 

        [STAThread]
        private void CopyToClipboard()
        {
            Clipboard.SetImage(ClipImage);
        }

        private void SetFont()
        {
            FontStyle ifontstyle = (FontStyle)iBold + iStrikeout + iItalic + iUnderline;
            ifont = new Font(ifontname, ifontsize, ifontstyle, GraphicsUnit.Point);
            rtb.SelectionFont = ifont;
        }
 
        private void SetFontDefaults()
        {
            iColor = Color.Black; // not really a font property
            ifontname = "Arial";       
            iBold = 0; //1
            iItalic = 0; //2
            iStrikeout = 0; // 8
            iUnderline = 0;// 4
            ifontsize = 10;
            SetFont();
        }

        // the printing bits
        //Convert the unit used by the .NET framework (1/100 inch) 
        //and the unit used by Win32 API calls (twips 1/1440 inch)
        private const double anInch = 14.4;

        [StructLayout(LayoutKind.Sequential)]
        private struct _RECT
        {
            public int Left;
            public int Top;
            public int Right;
            public int Bottom;
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct _CHARRANGE
        {
            public int cpMin;         //First character of range (0 for start of doc)
            public int cpMax;           //Last character of range (-1 for end of doc)
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct _FORMATRANGE
        {
            public IntPtr hdc;             //Actual DC to draw on
            public IntPtr hdcTarget;       //Target DC for determining text formatting
            public _RECT rc;               //Region of the DC to draw to (in twips)
            public _RECT rcPage;           //Region of the whole DC (page size) (in twips)
            public _CHARRANGE chrg;        //Range of text to draw (see earlier declaration)
        }

        private const int WM_USER = 0x0400;
        private const int EM_FORMATRANGE = WM_USER + 57;

        [DllImport("USER32.dll")]
        private static extern IntPtr SendMessage(IntPtr hWnd, int msg, IntPtr wp, IntPtr lp);

        // Render the contents of the RichTextBox for printing
        // Return the last character printed + 1 (printing start from this point for next page)
        private int print(int charFrom, int charTo, PrintPageEventArgs e)
        {
            //Calculate the area to render and print
            _RECT rectToPrint;
            rectToPrint.Top = (int)(e.MarginBounds.Top * anInch);
            rectToPrint.Bottom = (int)(e.MarginBounds.Bottom * anInch);
            rectToPrint.Left = (int)(e.MarginBounds.Left * anInch);
            rectToPrint.Right = (int)(e.MarginBounds.Right * anInch);

            //Calculate the size of the page
            _RECT rectPage;
            rectPage.Top = (int)(e.PageBounds.Top * anInch);
            rectPage.Bottom = (int)(e.PageBounds.Bottom * anInch);
            rectPage.Left = (int)(e.PageBounds.Left * anInch);
            rectPage.Right = (int)(e.PageBounds.Right * anInch);

            IntPtr hdc = e.Graphics.GetHdc();

            _FORMATRANGE fmtRange;
            fmtRange.chrg.cpMax = charTo;				//Indicate character from to character to 
            fmtRange.chrg.cpMin = charFrom;
            fmtRange.hdc = hdc;                    //Use the same DC for measuring and rendering
            fmtRange.hdcTarget = hdc;              //Point at printer hDC
            fmtRange.rc = rectToPrint;             //Indicate the area on page to print
            fmtRange.rcPage = rectPage;            //Indicate size of page

            IntPtr res = IntPtr.Zero;

            IntPtr wparam = IntPtr.Zero;
            wparam = new IntPtr(1);

            //Get the pointer to the FORMATRANGE structure in memory
            IntPtr lparam = IntPtr.Zero;
            lparam = Marshal.AllocCoTaskMem(Marshal.SizeOf(fmtRange));
            Marshal.StructureToPtr(fmtRange, lparam, false);

            //Send the rendered data for printing 
            res = SendMessage(rtb.Handle, EM_FORMATRANGE, wparam, lparam);

            //Free the block of memory allocated
            Marshal.FreeCoTaskMem(lparam);

            //Release the device context handle obtained by a previous call
            e.Graphics.ReleaseHdc(hdc);

            //Return last + 1 character printer
            return res.ToInt32();


        }

 
        // internal events concerned with printing
        private int checkPrint;

        private void pdoc_BeginPrint(object sender, System.Drawing.Printing.PrintEventArgs e)
        {
            checkPrint = 0;
        }

        private void pdoc_PrintPage(object sender, System.Drawing.Printing.PrintPageEventArgs e)
        {
            // Print the content of RichTextBox. Store the last character printed.
            checkPrint = print(checkPrint, rtb.TextLength, e);

            // Check for more pages
            if (checkPrint < rtb.TextLength)
                e.HasMorePages = true;
            else
                e.HasMorePages = false;
        }

        // When reading the documentation on EM_FORMATRANGE carefully, notice
        // that the value for lParam can also be NULL to "free information cached by the control", 
        // so it is a good idea to add another function to be called when printing has finished
        // (add an event handler for a PrintDocument's EndPrint event in the program):

        // Free cached data from rich edit control after printing
        //public void FormatRangeDone()
        //{
        //    IntPtr lParam = new IntPtr(0);
        //    SendMessage(Handle, EM_FORMATRANGE, 0, lParam);
        //}

    }



}
